
;------------------------------------------------------------------------------------------

pcint2_handler:
		push	r16
		in	r16,SREG		;backup SREG
		push	r16
		push	r0
		push	r1
		push	r2
		push	r3
		push	r4
		push	r5
		push	r6
		push	r7
		push	r8
		push	r9
		push	r10
		push	r11
		push	r12
		push	r13
		push	r14
		push	r15
		rcall do_rx
		pop	r15
		pop	r14
		pop	r13
		pop	r12
		pop	r11
		pop	r10
		pop	r9
		pop	r8
		pop	r7
		pop	r6
		pop	r5
		pop	r4
		pop	r3
		pop	r2
		pop	r1
		pop	r0
		ldi	r16,0x04		; prevent the interrupt handler from being called immediately again, as it would otherwise because since the start of the interrupt handler, the pin has changed many times
		out	PCIFR,r16 		; "the flag can be cleared by writing a logical 1 to it" - weird, but it works
		pop	r16
		out	SREG,r16		; restore SREG
		pop	r16
		reti



; subroutine called as soon as activity is seen on the ethernet input
; most likely, we're called very early in the preamble
; need to skip 8 bytes of preamble, 14 bytes of ethernet header, 20 bytes of IP header, 8 bytes of UDP header
; that's 50 bytes in total, or 400 bits, or 800 clock cycles
; but we've already spent a few clock cycles responding to the interrupt, and some 20 on instructions so far
; but that's fine: we want to take our 16 samples somewhere near the middle of the block of 128 that will be sent


do_rx:

		ldi r16,200
L06:		nop
		dec r16
		brne L06

		ldi xl,0


analoop:
							; take 16 consecutive samples
		in	r0,EthRxPort		; 16 clock cycles
		in	r1,EthRxPort
		in	r2,EthRxPort
		in	r3,EthRxPort
		in	r4,EthRxPort
		in	r5,EthRxPort
		in	r6,EthRxPort
		in	r7,EthRxPort
		in	r8,EthRxPort
		in	r9,EthRxPort
		in	r10,EthRxPort
		in	r11,EthRxPort
		in	r12,EthRxPort
		in	r13,EthRxPort
		in	r14,EthRxPort
		in	r15,EthRxPort

							; first check: are the samples not all the same?
							; if all the same, then line idle, frame has ended (or perhaps it was only a link pulse, not a frame)
							; we check by simply counting how many of the first 8 samples are 1s
		ldi	r16,0			; 19 clock cycles
		sbrc	r0,EthRxPin		
		inc	r16
		sbrc	r1,EthRxPin	
		inc	r16
		sbrc	r2,EthRxPin	
		inc	r16
		sbrc	r3,EthRxPin	
		inc	r16
		sbrc	r4,EthRxPin	
		inc	r16
		sbrc	r5,EthRxPin	
		inc	r16
		sbrc	r6,EthRxPin	
		inc	r16
		sbrc	r7,EthRxPin	
		inc	r16
		andi	r16,0x07			; if line idle, count was 0 or 8, so r16 is now 0
		breq	silent

							; add 8 odd-numbered samples, with alternating sign
							; this gives 0 or 8 if there's a 5 MHz signal
							; result goes into lower nibble of r16
		ldi	 r16,0			; 17 clock cycles
		sbrc	r0,EthRxPin
		inc	r16
		sbrs	r2,EthRxPin
		inc	r16
		sbrc	r4,EthRxPin
		inc	r16
		sbrs	r6,EthRxPin
		inc	r16
		sbrc	r8,EthRxPin
		inc	r16
		sbrs	r10,EthRxPin
		inc	r16
		sbrc	r12,EthRxPin
		inc	r16
		sbrs	r14,EthRxPin
		inc	r16

							; same for even-numbered samples
							; result goes into upper nibble of r16
		sbrc	r1,EthRxPin		; 16 clock cycles
		subi	r16,-16
		sbrs	r3,EthRxPin
		subi	r16,-16
		sbrc	r5,EthRxPin
		subi	r16,-16
		sbrs	r7,EthRxPin
		subi	r16,-16
		sbrc	r9,EthRxPin
		subi	r16,-16
		sbrs	r11,EthRxPin
		subi	r16,-16
		sbrc	r13,EthRxPin
		subi	r16,-16
		sbrs	r15,EthRxPin
		subi	r16,-16
					; now the two nibbles in r16 contain the scores for the even and odd bits
		st	x+,r16		; store in RAM; this instruction takes 2 cycles
		cpi	xl,200		; did we sample the signal often enough?
		breq	donesampling	; if so, we're finished sampling the signal
					; motivation for that 200: 200*64=12800 bits = 1600 bytes > maximum ethernet frame length, so frame has now definitely finished
					; note though that normally we'll already detect that the medium is idle before this

					; so far, we've used 16+19+17+16+2+1+1+2 cycles (including the rjmp at the end)
					; that's 74 cycles
					; we send one real user bit per 64 ethernet bits, or 128 clock cycles
					; so we have to waste another 128-74=54 cycles = 3*18

		ldi r16,18
L01:		dec r16
		brne L01

		rjmp analoop



silent:
;call toggleled
					; medium idle
		cpi	xl,20*9+1	; check whether we have enough samples?
		brcs	tooshort

donesampling:
		; now analyse the received data
		; this will happen byte-wise, in the sense of bytes of user data, each user data byte being sent as 9 blocks of 64 ethernetbits, and 9 stored bytes
		; we need to find out whether the odd or the even samples were good, corresponding to the lower and higher nibble in each stored byte
		; two criterions for that: all nibbles are 0 or 8, and the 0th and the 8th are the same (sync/ref), and serve as reference for a 0 data bit
		push r17
		push r18
		push r19
		push r20
		push yl
		ldi	yl,0
		ldi	xl,0

extractnextbyte:

		ld	r18,x 		; store 0th sample in r18
		ldi r17,0		; r17 will be OR of all samples, to check for any non-0/8
		mov 	r20,xl
		subi	r20,-10		; address of 8th sample
L05:
		ld	r16,x+		; loop over all samples to compute their OR
		or	r17,r16
		cp	xl,r20
		brne L05

		eor	r18,r16		; compare 0th and 8th sample
		andi	r18,0x88	; strip upper bit of that comparison
		ldi	r16,0x77	
		and	r16,r17		; check lower 3 bits of the OR of all samples, to check whether they were all 0/8
		or	r16,r18		; OR these results: if all goes is well, one nibble is 0, indicating which half of the data to use (even/odd samples)
		mov	r17,r16
		andi	r17,0x0f	; check lower nibble
		brne notlowernibble
		ldi	r18,0x08	; lower nibble ok: r18 will contain the mask to be used
		rjmp match
notlowernibble:
		andi	r16,0xf0	; check higher nibble
		brne nomatch
		ldi	r18,0x80	; higher nibble ok; r18 is mask
match:
		subi	xl,10		; move pointer back
		ld	r19,x+		; load r19 with the first sample, which will act as reference for what a '0' is
		ldi	r17,0		; r17 will contain the user data byte
		dec	r20		; r20 points just beyond the last sample to be used for this user data byte
L07:
		ld	r16,x+		; loop over all samples once more, this time to extract the databits and shift them into r17
		clc
		eor	r16,r19		; invert if needed
		and	r16,r18		; extract bit to be used (top of either lower or higher nibble)
		breq L08
		sec
L08:		ror	r17		; shift received bit into r17
		cp	xl,r20		; finished with this byte?
		brne L07

		st	y+,r17		; store result into RAM
		cpi	yl,20		; extracted all 20 bytes?
		brne	extractnextbyte
		call	rx_success	; yes, then do something with the received data

nomatch:
		pop yl
		pop r20
		pop r19
		pop r18
		pop r17
		
tooshort:
		cpi	xl,1		; check whether we have at least some samples, i.e., we've seen a real packet and not just a link pulse
		brcs	reallytooshort
					; short packet may have been an ARP request, so schedule an ARP reply transmission
		in r16,GPIOR1		; scheduling ARP transmissions is done through the GPIOR1 register, see the OVF2handler in avr_eth.asm
		andi r16,127		; basically: if the last ARP transmission was long ago (so GPIOR1=128), sent one very soon; otherwise, wait so there's some 1.6 seconds (128 timer interrupts) in between 
		out GPIOR1,r16
reallytooshort:
		ret


